import numpy as np

def count_adjacent_true(arr):
    assert len(arr.shape) == 1
    assert arr.dtype == np.bool
    if arr.size == 0:
        return np.empty(0, dtype=int), np.empty(0, dtype=int)
    sw = np.insert(arr[1:] ^ arr[:-1], [0, arr.shape[0]-1], values=True)
    swi = np.arange(sw.shape[0])[sw]
    offset = 0 if arr[0] else 1
    lengths = swi[offset+1::2] - swi[offset:-1:2]
    return swi[offset:-1:2], lengths



# -*- coding: utf-8 -*-
# -*- mode: python -*-
""" Core classes and functions for the quicksong library """
from typing import Sequence, Optional, Iterator, Tuple, Iterable, Union, Callable
import logging
import numpy as np

from .streaming import StreamingTransform

log = logging.getLogger("quicksong")
numeric = Union[int, float, np.number]
SpectrogramFeature = Callable[[np.ndarray], float]


def make_training_dataset(features, intervals, dt):
    """For each interval in intervals, extract the corresponding frames from features

    features - (t,k) array of features
    labels   - structured numpy array with fields "id", "start", and "stop"
    dt       - frame duration in the features array (same units as labels)

    Returns X (features array), y (labels)
    """
    out_feats = []
    out_labels = []
    for interval in intervals:
        id = interval["id"]
        frames = slice(int(interval["start"] / dt), int(interval["stop"] / dt))
        feats = features[frames, :]
        labels = np.ones(feats.shape[0], dtype="i") * id
        out_feats.append(feats)
        out_labels.append(labels)
    return np.row_stack(out_feats), np.concatenate(out_labels)


class FeatureExtractor:
    """Extracts features from acoustic signals.

    This class is basically glue for the streaming transformers that convert the
    input signal into a spectrogram and convolve the spectrogram features.
    Unlike these transformers, it is able to handle arbitrary-sized inputs.

    """

    def __init__(
        self,
        signal_transformer: StreamingTransform,
        features: Sequence[SpectrogramFeature],
        convolver: StreamingTransform,
    ):
        from copy import copy

        self.signal_t = signal_transformer
        self.frame_size = signal_transformer.block_size
        self.feature_t = features
        self.block_size = convolver.block_size
        self.convolvers = tuple(copy(convolver) for f in features)
        self.reset()

    def _calculate_features(self, frame: np.ndarray):
        stft = self.signal_t.process(frame)
        features = tuple(feat(stft) for feat in self.feature_t)
        self.feat_buf.append(features)
        self.feat_count += 1

    def reset(self):
        """Reset the signal and feature buffers"""
        from collections import deque

        self.signal_buf = np.zeros(self.signal_t.block_size)
        self.signal_buf_ptr = 0
        self.feat_buf = deque()
        self.feat_count = 0
        self.feat_prefilled = False

    def process(self, signal: np.ndarray):
        # calculate the spectrogram frames first and then store features.
        signal_ptr = 0
        if self.signal_buf_ptr > 0:
            needed = self.frame_size - self.signal_buf_ptr
            log.debug("filling incomplete frame with %ds samples", needed)
            self.signal_buf[self.signal_buf_ptr :] = signal[:needed]
            signal_ptr += needed
            self._calculate_features(self.signal_buf)
            self.signal_buf_ptr = 0
        for f, i in enumerate(range(signal_ptr, signal.size, self.frame_size)):
            s = signal[i : i + self.frame_size]
            log.debug("frame %d: sample %d, %d samples", f, i, s.size)
            if s.size == self.frame_size:
                self._calculate_features(s)
            else:
                # partial fill
                self.signal_buf[: s.size] = s
                self.signal_buf[s.size :] = 0
                self.signal_buf_ptr = s.size
        # process the feature buffer
        log.debug("processing feature buffer in blocks of %d frames", self.block_size)
        while self.feat_count > self.block_size:
            log.debug("buffer contains %d frames", self.feat_count)
            features = zip(*(self.feat_buf.popleft() for i in range(self.block_size)))
            self.feat_count -= self.block_size
            if not self.feat_prefilled:
                log.debug(" - prefilling the convolver")
                for feature, conv in zip(features, self.convolvers):
                    conv.prefill(np.asarray(feature))
                self.feat_prefilled = True
            else:
                log.debug(" - convolving")
                results = []
                for feature, conv in zip(features, self.convolvers):
                    results.append(conv.process(np.asarray(feature)))
                yield np.column_stack(results)

    def close(self) -> np.ndarray:
        """Process any remaining samples and return result"""
        assert self.feat_count < self.block_size, "too many features in the buffer!"
        if self.signal_buf_ptr > 0:
            log.debug("processing last frame with %d samples", self.signal_buf_ptr)
            self._calculate_features(self.signal_buf[: self.signal_buf_ptr])
        log.debug("feature buffer contains %d frames", self.feat_count)
        features = zip(*(self.feat_buf.popleft() for i in range(self.feat_count)))
        results = []
        for feature, conv in zip(features, self.convolvers):
            results.append(
                np.row_stack((conv.process(np.asarray(feature)), conv.close()))
            )
        self.reset()
        return np.column_stack(results)

    def process_all(self, signal: np.ndarray) -> np.ndarray:
        """Process an entire signal and then close the extractor"""
        out = []
        for block in self.process(signal):
            out.append(block)
        out.append(self.close())
        return np.row_stack(out)


class IntervalFinder:
    """Finds runs in a binary vector meeting duration criteria"""

    frame: int
    interval_start: Optional[int]
    interval_end: Optional[int]

    def __init__(self, max_gap: float, min_duration: float, dt: float):
        self.max_gap = int(max_gap / dt)
        self.min_duration = int(min_duration / dt)
        self.reset()

    def reset(self) -> None:
        self.frame = 0
        self.interval_start = None
        self.interval_end = None

    def process(self, frames: np.ndarray) -> Iterator[Tuple[int, int]]:
        """Find intervals of zeros in a binary vector, yielded as (start, stop) tuples"""
        for p in frames:
            if not p:
                if self.interval_start is None:
                    logging.debug("ON: %d", self.frame)
                    self.interval_start = self.frame
                self.interval_end = self.frame
            elif self.interval_end is None:
                pass
            elif self.frame - self.interval_end < self.max_gap:
                logging.debug("GAP: %d-%d", self.interval_end, self.frame)
                pass
            else:
                logging.debug(
                    "OFF: %d (duration=%d)",
                    self.interval_end,
                    self.interval_end - self.interval_start,
                )
                if self.interval_end - self.interval_start > self.min_duration:
                    yield (self.interval_start, self.interval_end)
                self.interval_end = self.interval_start = None
            self.frame += 1


def make_hanning_kernels(widths: Sequence[float]) -> np.ndarray:
    """Create an array of hanning kernels of specified widths. All the kernels
    are aligned such that tau=0 is in the center"""
    nk = len(widths)
    nt = max(widths)
    ht = nt // 2
    kernels = np.zeros((nt, nk))
    for i, ws in enumerate(widths):
        offset = ht - ws // 2
        w = np.hanning(ws)
        kernels[offset : offset + ws, i] = w / np.sum(w**2)
    return kernels


def pad_intervals(
    intervals: Iterable[Tuple[numeric, numeric]],
    pad_before: numeric,
    pad_after: numeric,
) -> Iterator[Tuple[numeric, numeric]]:
    it = ((s - pad_before, e + pad_after) for s, e in intervals)
    start, end = next(it)
    for s, e in it:
        if s <= end:
            end = e
        else:
            yield (start, end)
            start, end = s, e
    yield (start, end)



# -*- coding: utf-8 -*-
# -*- mode: python -*-
""" Module for doing streaming conversions """
from typing import Sequence
from abc import ABC, abstractmethod
import logging
import numpy as np

log = logging.getLogger("quicksong")


class StreamingTransform(ABC):
    """Abstract base class for streaming transforms"""

    @abstractmethod
    def reset(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def prefill(self, samples: np.ndarray) -> None:
        raise NotImplementedError

    @abstractmethod
    def process(self, block: np.ndarray) -> np.ndarray:
        raise NotImplementedError


class OverlapSaveConvolver(StreamingTransform):
    """Computes running convolution using the overlap-save method.

    This class is initialized with one or more convolution kernels. The signal
    to be convolved is then passed to the intialized object in blocks through
    the conv() method.

    This class assumes that the kernels are acausal, with tau=0 at the center.
    The block size is set to half the size of the largest kernel. In order for
    the convolution to be aligned with the time base of the input, the first
    block needs to be discarded and a block of zeros needs to be processed at
    the end. The prefill() and close() methods are provided for convenience.

    kernel:     the convolution kernel(s). For a single kernel, this is a 1-D array. For
                multiple kernels, this should be a (ntau, nkernel) 2-D array

    """

    block_size: int
    ntau: int
    nkernels: int
    nfft: int
    kernels: np.ndarray
    buffer: np.ndarray

    def __init__(self, kernel: np.ndarray):
        from math import ceil, log

        if kernel.ndim == 1:
            self.ntau = kernel.size
            self.nkernels = 1
        elif kernel.ndim == 2:
            self.ntau, self.nkernels = kernel.shape
        else:
            raise ValueError("Kernel input must be 1d or 2d")
        self.block_size = self.ntau // 2
        self.nfft = 2 ** ceil(log(self.block_size + self.ntau, 2))
        self.kernels = np.fft.rfft(kernel, n=self.nfft, axis=0)
        self.reset()

    def __copy__(self):
        """Clones the convolver with the same kernels and an empty buffer"""
        cls = self.__class__
        result = cls.__new__(cls)
        for attr in ("ntau", "nkernels", "block_size", "nfft"):
            result.__dict__[attr] = self.__dict__[attr]
        result.kernels = self.kernels
        result.reset()
        return result

    def reset(self) -> None:
        self.buffer = np.zeros(self.nfft)

    def prefill(self, signal: np.ndarray) -> None:
        """Prefill the buffer"""
        self.buffer[-signal.size:] = signal

    def process(self, block: np.ndarray) -> np.ndarray:
        """Process a block of samples. Check block_size property. Incomplete
        blocks are padded with zeros."""
        if block.ndim != 1:
            raise IncompatibleBlockSize("Input signal must be 1-D")
        elif block.size > self.block_size:
            raise IncompatibleBlockSize(
                f"Block size is too large (max {self.block_size}"
            )
        # shift the buffer and add the block to the end
        # self.buffer = np.roll(self.buffer, -self.block_size)
        self.buffer[: -block.size] = self.buffer[block.size:]
        self.buffer[-block.size:] = block
        S = np.fft.rfft(self.buffer, n=self.nfft)
        C = self.kernels * S[:, np.newaxis]
        c = np.fft.irfft(C, n=self.nfft, axis=0)
        return c[-block.size:]

    def close(self):
        """End the convolution by processing a block of zeros """
        return self.process(np.zeros(self.block_size))


class STFT(StreamingTransform):
    """Compute a running short-time fourier transform."""

    nfft: int
    window: np.ndarray
    block_size: int
    buffer: np.ndarray

    def __init__(
        self,
        sampling_rate: float,
        window: float,
        shift: float,
        frequency_range: Sequence[float],
    ):
        import libtfr
        from math import ceil, log

        window_samples = int(window * sampling_rate)
        self.nfft = 2 ** ceil(log(window_samples, 2))
        self.block_size = int(shift * sampling_rate)
        self.mfft = libtfr.mfft_precalc(self.nfft, np.hanning(window_samples))
        df = sampling_rate / self.nfft
        f = np.arange(0, sampling_rate, df)
        f1, f2 = frequency_range
        self.findx = ((f >= f1) & (f < f2)).nonzero()[0]
        self.reset()

    @property
    def prefill_size(self) -> int:
        """Recommended prefill size"""
        return self.nfft - self.block_size

    def reset(self) -> None:
        self.buffer = np.zeros(self.nfft)

    def prefill(self, signal: np.ndarray) -> None:
        """Prefill the buffer"""
        self.buffer[-signal.size :] = signal

    def process(self, block: np.ndarray) -> np.ndarray:
        if block.ndim != 1:
            raise IncompatibleBlockSize("Input signal must be 1-D")
        elif block.size < self.block_size:
            block = np.pad(block, (0, self.block_size - block.size), "constant", constant_values=0)
        elif block.size > self.block_size:
            raise IncompatibleBlockSize(
                f"Block size does not match spectrogram shift ({self.block_size})"
            )
        # shift the buffer and add the next frame to the end
        self.buffer[: -self.block_size] = self.buffer[self.block_size :]
        self.buffer[-self.block_size :] = block
        return self.mfft.mtfft(self.buffer)[self.findx, 0]


class IncompatibleBlockSize(ValueError):
    """Raised when data provided to convolver is not the right size or shape"""

    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# -*- mode: python -*-
from setuptools import setup, Extension
import numpy

setup(
    ext_modules=[
        Extension(
            "quicksong.features", sources=["quicksong/features.pyx"], language="c++"
        )
    ],
    include_dirs=[numpy.get_include()],
)



# -*- coding: utf-8 -*-
# -*- mode: python -*-
# distutils: language = c++
""" Module for calculating features from spectrograms """
cimport numpy as np
import numpy as np
import cython
from libc.math cimport log10, log, exp

cdef extern from "<complex.h>" namespace "std" nogil:
    double abs(double complex z)

@cython.boundscheck(False)
@cython.wraparound(False)
def total_power(np.ndarray [np.complex128_t, ndim=1] psd):
    cdef size_t i
    cdef size_t npoints = psd.size
    cdef double sum = 0
    for i in range(npoints):
        sum += abs(psd[i])
    return log10(sum)


@cython.boundscheck(False)
@cython.wraparound(False)
def wiener_entropy(np.ndarray [np.complex128_t, ndim=1] psd):
    cdef size_t i
    cdef size_t npoints = psd.size
    cdef double pow
    cdef double sum = 0
    cdef double logsum = 0
    for i in range(npoints):
        pow = abs(psd[i])
        sum += pow
        logsum += log(pow)
    return log10(exp(logsum / npoints) / (sum / npoints))




# -*- coding: utf-8 -*-
# -*- mode: python -*-
""" Example support vector classifier model """


def train_classifier(X, y, cls_params, split_params):
    from sklearn.svm import SVC
    from sklearn.model_selection import train_test_split
    from sklearn.pipeline import make_pipeline
    from sklearn.preprocessing import StandardScaler

    ppl = make_pipeline(StandardScaler(), SVC(**cls_params))

    X_train, X_test, y_train, y_test = train_test_split(X, y, **split_params)

    ppl.fit(X_train, y_train)
    score = ppl.score(X_test, y_test)
    return ppl, score


